#
# Copyright (C) 2010-2016 dtk contributors
#
# This file is part of the dtk project.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
require 'restclient'
require 'json'

module DTK
  module Common
    module ResponseTokens
      StatusOK = "ok"
      StatusNotok = "notok"
      DataField = "data"
      StatusField = "status"
      ErrorsField = "errors"
      ValidationField = "validation"
      ErrorsSubFieldCode = "code"
      ErrorsOriginalException = "original_exception"
      GenericError = "error"
    end

    module ResponseConstants
      # these trems dont get properly marshalled
      NilTerm = '***NIL***'
      BooleanTrue = '***TRUE***'
      BooleanFalse = '***FALSE***'
    end

    class Response < Hash
      include ResponseTokens

      module Term
        include ResponseConstants
        def self.nil()
          NilTerm
        end
        module Boolean
          def self.true()
            Term::BooleanTrue
          end
          def self.false()
            Term::BooleanFalse
          end
        end
      end

      def initialize(hash={})
        super()
        replace(hash)
      end
      def ok?()
        self[StatusField] == StatusOK
      end

      def validation_response?
        !self[ValidationField].nil?
      end

      def validation_message
        self[ValidationField]['message']
      end

      def error_message
        self["errors"] ? (self["errors"].map { |e| e["message"]}).join(', ') : nil
      end

      def validation_actions
        return self[ValidationField]['actions_needed']
      end

      def data(*data_keys)
        data = self[DataField]
        case data_keys.size
         when 0 then data||{}
         when 1 then data && data[internal_key_form(data_keys.first)]
         else data_keys.map{|key|data && data[internal_key_form(key)]}.compact
        end
      end

      def data_hash_form(*data_keys)
        ret = Hash.new
        unless data = self[DataField]
          return ret
        end

        if data_keys.size == 0
          data.inject(Hash.new){|h,(k,v)|h.merge(external_key_form(k) => v)}
        else
          data_keys.each do |k|
            if v = data[internal_key_form(k)]
              ret.merge!(external_key_form(k) => v)
            end
          end
          ret
        end
      end

      def set_data(*data_values)
        self[DataField]=data_values
      end

      def set_data_hash(data_hash)
        self[DataField]=data_hash
      end

      def data_ret_and_remove!(*data_keys)
        data = data()
        data_keys.map{|key|data.delete(internal_key_form(key))}
      end

      def add_data_value!(key,value)
        data()[key.to_s] = value
        self
      end

      def internal_key_form(key)
        key.to_s
      end
      def external_key_form(key)
        key.to_sym
      end
      private :internal_key_form,:external_key_form

      module ErrorMixin
        def ok?()
          false
        end
      end

      class Error < self
        include ErrorMixin
        def initialize(hash={})
          super(hash)
        end
      end

      class RestClientWrapper
        class << self
          include ResponseTokens
          def get_raw(url,body={},opts={},&block)
            error_handling(opts) do
              url_with_params = generate_query_params_url(url, body)
              raw_response = ::RestClient::Resource.new(url_with_params,opts).get()
              block ? block.call(raw_response) : raw_response
            end
          end

          def get(url, body={}, opts={})
            get_raw(url,body, opts){|raw_response|Response.new(json_parse_if_needed(raw_response))}
          end

          def post_raw(url,body={},opts={},&block)
            error_handling(opts) do
              raw_response = ::RestClient::Resource.new(url,opts).post(body)
              block ? block.call(raw_response) : raw_response
            end
          end

          def delete_raw(url,body={},opts={},&block)
            error_handling(opts) do
              # DELETE method supports only query params
              url_with_params = generate_query_params_url(url, body)
              raw_response = ::RestClient::Resource.new(url_with_params,opts).delete()
              block ? block.call(raw_response) : raw_response
            end
          end

          def post(url,body={},opts={})
            post_raw(url,body,opts){|raw_response|Response.new(json_parse_if_needed(raw_response))}
          end

          def delete(url, body={}, opts={})
            delete_raw(url,body,opts){|raw_response|Response.new(json_parse_if_needed(raw_response))}
          end

          def json_parse_if_needed(item)
            item.kind_of?(String) ? JSON.parse(item) : item
          end
         private

          def generate_query_params_url(url, params_hash)
            if params_hash.empty?
              return url
            else
              query_params_string = params_hash.map { |k,v| "#{CGI.escape(k.to_s)}=#{CGI.escape(v.to_s)}" }.join('&')
              return url.concat('?').concat(query_params_string)
            end
          end

          def safe_json_parse(string)
            begin
              JSON.parse(string)
            rescue => e
              nil
            end
          end

          def errors_field(msg=nil)
            [msg.nil? ? {} : {'message' => msg}]
          end

          def error_handling(opts={},&block)
            begin
              block.call
            rescue  ::RestClient::ResourceNotFound, RestClient::Request::Unauthorized, RestClient::BadRequest,::RestClient::InternalServerError => e
              # with latest set of changes we will consider this as special case since most of legacy code is expecting Response class
              parsed_response = safe_json_parse(e.response) || {}
              errors = parsed_response['errors'] || parsed_response['error'] || errors_field('Server Error')
              Response.new(StatusField => StatusNotok, ErrorsField => errors)
            rescue ::RestClient::Forbidden => e
              return error_response({ErrorsSubFieldCode => RestClientErrors[e.class.to_s]||GenericError, ErrorsOriginalException => e},opts) unless e.inspect.to_s.include?("PG::Error")

              errors = {"code" => "pg_error", "message" => e.inspect.to_s.strip, ErrorsOriginalException => e}
              error_response(errors)
            rescue ::RestClient::ServerBrokeConnection,::RestClient::RequestTimeout, Errno::ECONNREFUSED => e
              error_response({ErrorsSubFieldCode => RestClientErrors[e.class.to_s]||GenericError, ErrorsOriginalException => e},opts)
            rescue Exception => e
              error_response({ErrorsSubFieldCode => RestClientErrors[e.class.to_s], ErrorsOriginalException => e},opts)
            end
          end

          def error_response(error_or_errors,opts={})
            errors = error_or_errors.kind_of?(Hash) ? [error_or_errors] : error_or_errors
            (opts[:error_response_class]||Error).new(StatusField => StatusNotok, ErrorsField => errors)
          end

          RestClientErrors = {
            "RestClient::Forbidden" => "forbidden",
            "RestClient::ServerBrokeConnection" => "broken",
            "RestClient::Request::Unauthorized" => "unauthorized",
            "RestClient::ServiceUnavailable" => "unavailable",
            "RestClient::InternalServerError" => "internal_server_error",
            "RestClient::RequestTimeout" => "timeout",
            "RestClient::ResourceNotFound" => "resource_not_found",
            "Errno::ECONNREFUSED" => "connection_refused"
          }
        end
      end
    end
  end
end
